# Author: Eric Bezzam
# Date: July 15, 2016

from .music import *
from scipy.linalg import svdvals

from scipy import linalg

class TOPS(MUSIC):
    """
    Class to apply Test of Orthogonality of Projected Subspaces [TOPS]_ for 
    Direction of Arrival (DoA) estimation.

    .. note:: Run locate_source() to apply the TOPS algorithm.

    Parameters
    ----------
    L: numpy array
        Microphone array positions. Each column should correspond to the 
        cartesian coordinates of a single microphone.
    fs: float
        Sampling frequency.
    nfft: int
        FFT length.
    c: float
        Speed of sound. Default: 343 m/s
    num_src: int
        Number of sources to detect. Default: 1
    mode: str
        'far' or 'near' for far-field or near-field detection 
        respectively. Default: 'far'
    r: numpy array
        Candidate distances from the origin. Default: np.ones(1)
    azimuth: numpy array
        Candidate azimuth angles (in radians) with respect to x-axis.
        Default: np.linspace(-180.,180.,30)*np.pi/180
    colatitude: numpy array
        Candidate elevation angles (in radians) with respect to z-axis.
        Default is x-y plane search: np.pi/2*np.ones(1)

    References
    ----------

    .. [TOPS] Y. Yeo-Sun, L. M. Kaplan, J. H. McClellan, *TOPS: New DOA estimator for wideband 
        signals*, IEEE Trans. Signal Process., Vol. 54, Num 6., pp 1977--1989, 2006

    """
    def __init__(self, L, fs, nfft, c=343.0, num_src=1, mode='far', r=None,
        azimuth=None, colatitude=None, **kwargs):

        MUSIC.__init__(self, L=L, fs=fs, nfft=nfft, c=c, num_src=num_src, 
            mode=mode, r=r, azimuth=azimuth, colatitude=colatitude, **kwargs)

    def _process(self, X):
        """
        Perform TOPS for a given frame in order to estimate steered response 
        spectrum.
        """

        # need more than 1 frequency band
        if self.num_freq < 2:
            raise ValueError('Need more than one frequency band!')

        # select reference frequency (largest power)
        max_bin = np.argmax(np.sum(np.sum(abs(X[:,self.freq_bins,:]),axis=0),
            axis=1))
        f0 = self.freq_bins[max_bin]
        freq = list(self.freq_bins)
        freq.remove(f0)

        # compute empirical cross correlation matrices
        C_hat = self._compute_correlation_matrices(X)

        # compute signal and noise subspace for each frequency band
        F = np.zeros((self.num_freq,self.M,self.num_src), dtype='complex64')
        W = np.zeros((self.num_freq,self.M,self.M-self.num_src), 
            dtype='complex64')
        for k in range(self.num_freq):
            # subspace decomposition
            F[k,:,:], W[k,:,:], ws, wn = \
                self._subspace_decomposition(C_hat[k,:,:])

        # create transformation matrix
        f = 1.0/self.nfft/self.c*1j*2*np.pi*self.fs*(np.linspace(0,self.nfft/2,
            self.nfft/2+1)-f0)

        Phi = np.zeros((len(f),self.M,self.grid.n_points), dtype='complex64')

        for n in range(self.grid.n_points):
            p_s = self.grid.cartesian[:self.grid.dim,n]
            proj = np.dot(p_s,self.L[:self.grid.dim,:])
            for m in range(self.M):
                Phi[:,m,n] = np.exp(f*proj[m])

        # determine direction using rank test
        for n in range(self.grid.n_points):
            # form D matrix
            D = np.zeros((self.num_src,(self.M-self.num_src)*
                (self.num_freq-1)), dtype='complex64')
            for k in range(self.num_freq-1):
                Uk = np.conjugate(np.dot(np.diag(Phi[k,:,n]), 
                    F[max_bin,:,:])).T
                    # F[max_bin,:,:])).T
                idx = range(k*(self.M-self.num_src),(k+1)*(self.M-self.num_src))
                D[:,idx] = np.dot(Uk,W[k,:,:])
            s = svdvals(D)
            self.grid.values[n] = 1.0/s[-1]




# Author: Eric Bezzam
# Date: Feb 1, 2016

"""Class for real-time STFT analysis and processing."""
from __future__ import division

import numpy as np
from .dft import DFT
import warnings

try:
    import matplotlib as mpl
    matplotlib_available = True
except ImportError:
    matplotlib_available = False

if matplotlib_available:
    import matplotlib.pyplot as plt

class STFT(object):
    """
    A class STFT processing.

    Parameters
    -----------
    N : int
        number of samples per frame
    fs : float
        Sampling frequency.
    hop : int
        hop size
    analysis_window : numpy array
        window applied to block before analysis
    synthesis : numpy array
        window applied to the block before synthesis
    channels : int
        number of signals
    transform: str, optional
        which FFT package to use: 'numpy', 'pyfftw', or 'mkl'
    """
    def __init__(self, N, fs, hop=None, analysis_window=None, 
        synthesis_window=None, channels=1, transform='numpy'):
        # initialize parameters
        self.N = N          # number of samples per frame
        self.fs = fs
        self.D = channels         # number of channels
        if hop is not None: # hop size
            self.hop = hop  
        else:
            self.hop = self.N/2
        self.hop = int(np.floor(self.hop))

        # analysis window
        if analysis_window is not None:
            self.analysis_window = analysis_window
        elif analysis_window is None and self.hop ==self.N/2:
            self.analysis_window = np.hanning(self.N)
        else:
            self.analysis_window = None
        # synthesis window
        if synthesis_window is not None:  
            self.synthesis_window = synthesis_window
        elif synthesis_window is None and self.hop ==self.N/2:
            self.synthesis_window = None # rectangular window
        else:
            self.synthesis_window = None

        # create DFT object
        self.transform = transform
        self.nfft = self.N
        self.nbin = self.nfft // 2 + 1
        self.freq = np.linspace(0,self.fs/2,self.nbin)
        self.dft = DFT(nfft=self.nfft,fs=self.fs,D=self.D,
                analysis_window=self.analysis_window,
                synthesis_window=self.synthesis_window,
                transform=self.transform)

        self.fft_out_buffer = np.zeros(self.nbin, dtype=np.complex64)

        # initialize filter + zero padding --> use set_filter
        self.zf = 0; self.zb = 0
        self.H = None       # filter frequency spectrum

        # state variables
        self.num_frames = 0            # number of frames processed so far
        self.n_state = self.N - self.hop

        # allocate all the required buffers
        self._make_buffers()

    def _make_buffers(self):

        if self.D==1:  # need this distinction for fftw

            # The input buffer, float32 for speed!
            self.fft_in_buffer = np.zeros(self.nfft, dtype=np.float32)
            #  a number of useful views on the input buffer
            self.fft_in_state = self.fft_in_buffer[self.zf:self.zf+self.n_state]  # Location of state
            self.fresh_samples = self.fft_in_buffer[self.zf+self.n_state:self.zf+self.n_state+self.hop]
            self.old_samples = self.fft_in_buffer[self.zf+self.hop:self.zf+self.hop+self.n_state]

            self.x_p = np.zeros(self.n_state, dtype=np.float32)  # State buffer
            self.y_p = np.zeros(self.nfft - self.hop, dtype=np.float32)  # prev reconstructed samples
            self.X = np.zeros(self.nbin, dtype=np.complex64)       # current frame in STFT domain
            self.out = np.zeros(self.hop, dtype=np.float32)

        else:

            # The input buffer, float32 for speed!
            self.fft_in_buffer = np.zeros((self.nfft, self.D), dtype=np.float32)
            #  a number of useful views on the input buffer
            self.fft_in_state = self.fft_in_buffer[self.zf:self.zf+self.n_state,:]  # Location of state
            self.fresh_samples = self.fft_in_buffer[self.zf+self.n_state:self.zf+self.n_state+self.hop,:]
            self.old_samples = self.fft_in_buffer[self.zf+self.hop:self.zf+self.hop+self.n_state,:]

            self.x_p = np.zeros((self.n_state, self.D), dtype=np.float32)  # State buffer
            self.y_p = np.zeros((self.nfft - self.hop, self.D), dtype=np.float32)  # prev reconstructed samples
            self.X = np.zeros((self.nbin, self.D), dtype=np.complex64)       # current frame in STFT domain
            self.out = np.zeros((self.hop,self.D), dtype=np.float32)


    def reset(self):
        """
        Reset state variables. Necesary after changing or setting the filter or zero padding.
        """
        self.num_frames = 0
        self.nbin = self.nfft // 2 + 1
        self.freq = np.linspace(0,self.fs/2,self.nbin)

        if self.D==1:
            self.fft_in_buffer[:] = 0.
            self.X[:] = 0.
            self.y_p[:] = 0.
        else:
            self.fft_in_buffer[:,:] = 0.
            self.X[:,:] = 0.
            self.y_p[:,:] = 0.

        self.dft = DFT(nfft=self.nfft,fs=self.fs,D=self.D,
            analysis_window=self.analysis_window,
            synthesis_window=self.synthesis_window,
            transform=self.transform)

    def zero_pad_front(self, zf):
        """
        Set zero-padding at beginning of frame.
        """
        self.zf = zf
        self.nfft = self.N+self.zb+self.zf
        if self.analysis_window is not None:
            self.analysis_window = np.concatenate((np.zeros(zf), self.analysis_window))
        if self.synthesis_window is not None:
            self.synthesis_window = np.concatenate((np.zeros(zf), self.synthesis_window))

    def zero_pad_back(self, zb):
        """
        Set zero-padding at end of frame.
        """
        self.zb = zb
        self.nfft = self.N+self.zb+self.zf
        if self.analysis_window is not None:
            self.analysis_window = np.concatenate((self.analysis_window, np.zeros(zb)))
        if self.synthesis_window is not None:
            self.synthesis_window = np.concatenate((self.synthesis_window, np.zeros(zb)))

    def set_filter(self, coeff, zb=None, zf=None, freq=False):
        """
        Set time-domain filter with appropriate zero-padding.
        Frequency spectrum of the filter is computed and set for the object. 
        There is also a check for sufficient zero-padding.

        Parameters
        -----------
        coeff : numpy array 
            Filter in time domain.
        zb : int
            Amount of zero-padding added to back/end of frame.
        zf : int
            Amount of zero-padding added to front/beginning of frame.
        """
        # apply zero-padding
        if zb is not None:
            self.zero_pad_back(zb)
        if zf is not None:
            self.zero_pad_front(zf)  
        self.reset()      
        if not freq:
            # compute filter magnitude and phase spectrum
            self.H = np.complex64(np.fft.rfft(coeff, self.nfft, axis=0))
            # check for sufficient zero-padding
            if self.nfft < (self.N+len(coeff)-1):
                raise ValueError('Insufficient zero-padding for chosen number of samples per frame (L) and filter length (h). Require zero-padding such that new length is at least (L+h-1).')
        else:
            if len(coeff)!=self.nbin:
                raise ValueError('Invalid length for frequency domain coefficients.')
            self.H = coeff

        # We need to reallocate buffers after changing zero padding
        self._make_buffers()


    def analysis(self, x_n):
        """
        Transform new samples to STFT domain for analysis.
        Parameters
        -----------
        x_n : numpy array
            [self.hop] new samples.
        Returns
        -----------
        self.X : numpy array 
            Frequency spectrum of given frame.
        """

        # check for valid input - already done by self.dft
        # if x_n.shape[0]!=self.hop:
        #     raise ValueError('Invalid input dimensions.')
        # if self.D > 1 and x_n.shape[1]!=self.D:
        #     raise ValueError('Invalid input dimensions.')

        self.fresh_samples[:,] = x_n[:,]  # introduce new samples

        self.x_p[:,] = self.old_samples   # save next state

        # apply DFT to current frame
        self.X[:] = self.dft.analysis(self.fft_in_buffer)

        # shift backwards in the buffer the state
        self.fft_in_state[:,] = self.x_p[:,]

        # self.num_frames += 1

    def process(self):
        """
        Apply filtering in STFT domain.
        Returns
        -----------
        self.X : numpy array 
            Frequency spectrum of given frame.
        """
        if self.H is not None:
            np.multiply(self.X, self.H, self.X)


    def synthesis(self):
        """
        Transform to time domain and reconstruct output with overlap-and-add.

        Returns
        -------
        numpy array
            Reconstructed array of samples of length [self.hop]
        """
        # apply IDFT to current frame
        self.dft.synthesis(self.X)

        # reconstruct output
        L = self.y_p.shape[0]  # length of output state vector

        self.out[:,] = self.dft.x[0:self.hop,]  # fresh output samples

        # add state from previous frames when overlap is used
        if L > 0:
            m = np.minimum(self.hop, L)
            self.out[:m,] += self.y_p[:m,]
            # update state variables
            self.y_p[:-self.hop,] = self.y_p[self.hop:,]  # shift out left
            self.y_p[-self.hop:,] = 0.
            self.y_p[:,] += self.dft.x[-L:,]

        return self.out


    def get_prev_samples(self):
        """
        Get reconstructed previous samples.
        """
        return self.y_p



'''
Adaptive Filter Algorithms
==========================

This sub-package provides implementations of popular adaptive filter algorithms.

RLS
    | Recursive Least Squares
    | :py:obj:`pyroomacoustics.adaptive.rls` 
LMS
    | Least Mean Squares and Normalized Least Mean Squares
    | :py:obj:`pyroomacoustics.adaptive.lms`

All these classes derive from the base class
:py:obj:`pyroomacoustics.adaptive.adaptive_filter.AdaptiveFilter` that offer
a generic way of running an adaptive filter.

How to use the adaptive filter module
-------------------------------------

First, an adaptive filter object is created and all the relevant options
can be set (step size, regularization, etc). Then, the update function
is repeatedly called to provide new samples to the algorithm.

::

    # initialize the filter
    rls = pyroomacoustics.adaptive.RLS(30)

    # run the filter on a stream of samples
    for i in range(100):
        rls.update(x[i], d[i])

    # the reconstructed filter is available
    print('Reconstructed filter:', rls.w)


Other Available Subpackages
---------------------------

:py:obj:`pyroomacoustics.adaptive.data_structures`
    this provides abstractions for computing functions on regular or irregular
    grids defined on circles and spheres with peak finding methods

:py:obj:`pyroomacoustics.adaptive.util`
    a few methods mainly to efficiently manipulate Toeplitz and Hankel matrices

Utilities
---------

:py:obj:`pyroomacoustics.adaptive.algorithms`
    a dictionary containing all the adaptive filter object subclasses availables indexed by
    keys ``['RLS', 'BlockRLS', 'BlockLMS', 'NLMS']``

'''

from .adaptive_filter import *
from .lms import *
from .rls import *
from .util import *
from .data_structures import *

# Create this dictionary as a shortcut to different algorithms
algorithms = {
        'RLS' : RLS,
        'BlockRLS' : BlockRLS,
        'NLMS' : NLMS,
        'BlockLMS' : BlockLMS,
        }




# Author: Eric Bezzam
# Date: July 15, 2016

from .doa import *

class MUSIC(DOA):
    """
    Class to apply MUltiple SIgnal Classication (MUSIC) direction-of-arrival 
    (DoA) for a particular microphone array.

    .. note:: Run locate_source() to apply the MUSIC algorithm.

    Parameters
    ----------
    L: numpy array
        Microphone array positions. Each column should correspond to the 
        cartesian coordinates of a single microphone.
    fs: float
        Sampling frequency.
    nfft: int
        FFT length.
    c: float
        Speed of sound. Default: 343 m/s
    num_src: int
        Number of sources to detect. Default: 1
    mode: str
        'far' or 'near' for far-field or near-field detection 
        respectively. Default: 'far'
    r: numpy array
        Candidate distances from the origin. Default: np.ones(1)
    azimuth: numpy array
        Candidate azimuth angles (in radians) with respect to x-axis.
        Default: np.linspace(-180.,180.,30)*np.pi/180
    colatitude: numpy array
        Candidate elevation angles (in radians) with respect to z-axis.
        Default is x-y plane search: np.pi/2*np.ones(1)
    """
    def __init__(self, L, fs, nfft, c=343.0, num_src=1, mode='far', r=None,
        azimuth=None, colatitude=None, **kwargs):

        DOA.__init__(self, L=L, fs=fs, nfft=nfft, c=c, num_src=num_src, 
            mode=mode, r=r, azimuth=azimuth, colatitude=colatitude, **kwargs)

        self.Pssl = None

    def _process(self, X):
        """
        Perform MUSIC for given frame in order to estimate steered response 
        spectrum.
        """

        # compute steered response
        self.Pssl = np.zeros((self.num_freq,self.grid.n_points))
        num_freq = self.num_freq

        C_hat = self._compute_correlation_matrices(X)

        for i in range(self.num_freq):
            k = self.freq_bins[i]

            # subspace decomposition
            Es, En, ws, wn = self._subspace_decomposition(C_hat[i,:,:])

            # compute spatial spectrum
            # cross = np.dot(En,np.conjugate(En).T)
            cross = np.identity(self.M) - np.dot(Es, np.conjugate(Es).T) 
            self.Pssl[i,:] = self._compute_spatial_spectrum(cross,k)

        self.grid.set_values(np.sum(self.Pssl, axis=0)/num_freq)

    def plot_individual_spectrum(self):
        """
        Plot the steered response for each frequency.
        """

        # check if matplotlib imported
        if matplotlib_available is False:
            warnings.warn('Could not import matplotlib.')
            return

        # only for 2D
        if self.grid.dim == 3:
            pass
        else:
            warnings.warn('Only for 2D.')
            return

        # plot
        for k in range(self.num_freq):

            freq = float(self.freq_bins[k])/self.nfft*self.fs
            azimuth = self.grid.azimuth * 180 / np.pi

            plt.plot(azimuth, self.Pssl[k,0:len(azimuth)])

            plt.ylabel('Magnitude')
            plt.xlabel('Azimuth [degrees]')
            plt.xlim(min(azimuth),max(azimuth))
            plt.title('Steering Response Spectrum - ' + str(freq) + ' Hz')
            plt.grid(True)

    def _compute_spatial_spectrum(self,cross,k):

        P = np.zeros(self.grid.n_points)

        for n in range(self.grid.n_points):
            Dc = np.array(self.mode_vec[k,:,n],ndmin=2).T
            Dc_H = np.conjugate(np.array(self.mode_vec[k,:,n],ndmin=2))
            denom = np.dot(np.dot(Dc_H,cross),Dc)
            P[n] = 1/abs(denom)

        return P

    def _compute_correlation_matrices(self, X):
        C_hat = np.zeros([self.num_freq,self.M,self.M], dtype=complex)
        for i in range(self.num_freq):
            k = self.freq_bins[i]
            for s in range(self.num_snap):
                C_hat[i,:,:] = C_hat[i,:,:] + np.outer(X[:,k,s], 
                    np.conjugate(X[:,k,s]))
        return C_hat/self.num_snap

    def _subspace_decomposition(self, R):

        # eigenvalue decomposition!
        w,v = np.linalg.eig(R)

        # sort out signal and noise subspace
        # Signal comprises the leading eigenvalues
        # Noise takes the rest
        eig_order = np.flipud(np.argsort(abs(w)))
        sig_space = eig_order[:self.num_src]
        noise_space = eig_order[self.num_src:]

        # eigenvalues
        ws = w[sig_space]
        wn = w[noise_space]

        # eigenvectors
        Es = v[:,sig_space]
        En = v[:,noise_space]

        return Es, En, ws, wn




# @version: 1.0  date: 05/06/2015 by Sidney Barthe
# @author: robin.scheibler@epfl.ch, ivan.dokmanic@epfl.ch, sidney.barthe@epfl.ch
# @copyright: EPFL-IC-LCAV 2015

'''A collection of windowing functions.'''

import numpy as np

# cosine window function
def cosine(N, flag='asymmetric', length='full'):
    '''
    The cosine window function

    .. math:: 
    
        w[n] = \cos(\pi (n/M - 0.5))^2

    Parameters
    ----------
    N: int
        the window length
    flag: string, optional
        Possible values

        - *asymmetric*: asymmetric windows are used for overlapping transforms (:math:`M=N`)
        - *symmetric*: the window is symmetric (:math:`M=N-1`)
        - *mdct*: impose MDCT condition on the window (:math:`M=N-1` and
          :math:`w[n]^2 + w[n+N/2]^2=1`)

    length: string, optional
        Possible values

        - *full*: the full length window is computed
        - *right*: the right half of the window is computed
        - *left*: the left half of the window is computed
    '''

    # first choose the indexes of points to compute
    if (length == 'left'):     # left side of window
        t = np.arange(0, N / 2)
    elif(length == 'right'):   # right side of window
        t = np.arange(N / 2, N)
    else:                   # full window by default
        t = np.arange(0, N)

    # if asymmetric window, denominator is N, if symmetric it is N-1
    if (flag == 'symmetric' or flag == 'mdct'):
        t = t / float(N - 1)
    else:
        t = t / float(N)

    w = np.cos(np.pi * (t - 0.5)) ** 2

    # make the window respect MDCT condition
    if (flag == 'mdct'):
        w **= 2
        d = w[:N / 2] + w[N / 2:]
        w[:N / 2] *= 1. / d
        w[N / 2:] *= 1. / d

    # compute window
    return w


# triangular window function
def triang(N, flag='asymmetric', length='full'):
    '''
    The triangular window function

    .. math:: 
    
        w[n] = 1 - | 2 n / M - 1 |, n=0,\ldots,N-1

    Parameters
    ----------
    N: int
        the window length
    flag: string, optional
        Possible values

        - *asymmetric*: asymmetric windows are used for overlapping transforms (:math:`M=N`)
        - *symmetric*: the window is symmetric (:math:`M=N-1`)
        - *mdct*: impose MDCT condition on the window (:math:`M=N-1` and
          :math:`w[n]^2 + w[n+N/2]^2=1`)

    length: string, optional
        Possible values

        - *full*: the full length window is computed
        - *right*: the right half of the window is computed
        - *left*: the left half of the window is computed
    '''


    # first choose the indexes of points to compute
    if (length == 'left'):     # left side of window
        t = np.arange(0, N / 2)
    elif(length == 'right'):   # right side of window
        t = np.arange(N / 2, N)
    else:                   # full window by default
        t = np.arange(0, N)

    # if asymmetric window, denominator is N, if symmetric it is N-1
    if (flag == 'symmetric' or flag == 'mdct'):
        t = t / float(N - 1)
    else:
        t = t / float(N)

    w = 1. - np.abs(2. * t - 1.)

    # make the window respect MDCT condition
    if (flag == 'mdct'):
        d = w[:N / 2] + w[N / 2:]
        w[:N / 2] *= 1. / d
        w[N / 2:] *= 1. / d

    # compute window
    return w


# hann window function
def hann(N, flag='asymmetric', length='full'):
    '''
    The Hann window function

    .. math:: 
        
        w[n] = 0.5 (1 - \cos(2 \pi n / M)), n=0,\ldots,N-1

    Parameters
    ----------
    N: int
        the window length
    flag: string, optional
        Possible values

        - *asymmetric*: asymmetric windows are used for overlapping transforms (:math:`M=N`)
        - *symmetric*: the window is symmetric (:math:`M=N-1`)
        - *mdct*: impose MDCT condition on the window (:math:`M=N-1` and
          :math:`w[n]^2 + w[n+N/2]^2=1`)

    length: string, optional
        Possible values

        - *full*: the full length window is computed
        - *right*: the right half of the window is computed
        - *left*: the left half of the window is computed
    '''


    # first choose the indexes of points to compute
    if (length == 'left'):     # left side of window
        t = np.arange(0, N / 2)
    elif(length == 'right'):   # right side of window
        t = np.arange(N / 2, N)
    else:                   # full window by default
        t = np.arange(0, N)

    # if asymmetric window, denominator is N, if symmetric it is N-1
    if (flag == 'symmetric' or flag == 'mdct'):
        t = t / float(N - 1)
    else:
        t = t / float(N)

    w = 0.5 * (1 - np.cos(2 * np.pi * t))

    # make the window respect MDCT condition
    if (flag == 'mdct'):
        d = w[:N / 2] + w[N / 2:]
        w[:N / 2] *= 1. / d
        w[N / 2:] *= 1. / d

    # compute window
    return w


# Blackman-Harris window
def blackman_harris(N, flag='asymmetric', length='full'):
    '''
    The Hann window function

    .. math:: 
    
        w[n] = a_0 - a_1 \cos(2\pi n/M) + a_2 \cos(4\pi n/M) + a_3 \cos(6\pi n/M), n=0,\ldots,N-1

    Parameters
    ----------
    N: int
        the window length
    flag: string, optional
        Possible values

        - *asymmetric*: asymmetric windows are used for overlapping transforms (:math:`M=N`)
        - *symmetric*: the window is symmetric (:math:`M=N-1`)

    length: string, optional
        Possible values

        - *full*: the full length window is computed
        - *right*: the right half of the window is computed
        - *left*: the left half of the window is computed
    '''


    # coefficients
    a = np.array([.35875, .48829, .14128, .01168])

    # first choose the indexes of points to compute
    if (length == 'left'):     # left side of window
        t = np.arange(0, N / 2)
    elif(length == 'right'):   # right side of window
        t = np.arange(N / 2, N)
    else:                   # full window by default
        t = np.arange(0, N)

    # if asymmetric window, denominator is N, if symmetric it is N-1
    if (flag == 'symmetric'):
        t = t / float(N - 1)
    else:
        t = t / float(N)

    pi = np.pi
    w = a[0] - a[1]*np.cos(2*pi*t) + a[2]*np.cos(4*pi*t) + a[3]*np.cos(6*pi*t)

    return w

# Rectangular window function
def rect(N):
    '''
    The rectangular window

    .. math:: 
    
        w[n] = 1, n=0,\ldots,N-1

    Parameters
    ----------
    N: int
        the window length
    '''

    return np.ones(N)



# Author: Eric Bezzam
# Date: Jan 31, 2016

"""
Methods for using the Discrete Fourier Transform (DFT) for the analysis and
synthesis of real signals.

When available, it is able to use pyfftw package or mkl fft.
"""

import numpy as np
from numpy.fft import rfft
from numpy.fft import irfft
import warnings

try:
    import matplotlib as mpl
    matplotlib_available = True
except ImportError:
    matplotlib_available = False

try:
    import pyfftw
    pyfftw_available = True
except ImportError:
    pyfftw_available = False

try:
    import mkl_fft  # https://github.com/LCAV/mkl_fft
    mkl_available = True
except ImportError:
    mkl_available = False

if matplotlib_available:
    import matplotlib.pyplot as plt

class DFT(object):
    """
    Parent class for performing Fourier Analysis of real signals through the
    Discrete Fourier Transform (DFT).

    Parameters
    ----------
    nfft: int
        FFT size.
    D: int
        Number of channels. Default is 1.
    fs: float
        Sampling frequency.
    analysis_window: numpy array
        Window to be applied before DFT.
    synthesis_window: numpy array
        Window to be applied after inverse DFT.
    transform: str
        'fftw' or 'numpy' to use the appropriate library..

    """

    def __init__(self, nfft, D=1, fs=1.0, analysis_window=None, synthesis_window=None, transform='numpy'):

        self.nfft = nfft
        self.D = D
        
        self.nbin = self.nfft//2+1
        self.fs = float(fs)
        self.freq = np.linspace(0,self.fs/2,self.nbin)
        self.X = np.squeeze(np.zeros((self.nbin,self.D),dtype='complex64'))
        self.x = np.squeeze(np.zeros((self.nfft,self.D),dtype='float32'))

        if analysis_window is not None:
            if self.D==1:
                self.analysis_window = analysis_window.astype('float32')
            else:
                self.analysis_window = np.tile(analysis_window, (self.D,1)).astype('float32').T
        else:
            self.analysis_window = None
        if synthesis_window is not None:
            if self.D==1:
                self.synthesis_window = synthesis_window.astype('float32')
            else:
                self.synthesis_window = np.tile(synthesis_window, (self.D,1)).astype('float32').T
        else:
            self.synthesis_window=None


        if transform == 'fftw':
            if pyfftw_available:
                import pyfftw
                self.transform = transform
                # allocate input and output (assuming real input) for pyfftw
                if self.D==1:
                    self.a = pyfftw.empty_aligned(self.nfft, dtype='float32')
                    self.b = pyfftw.empty_aligned(self.nbin, dtype='complex64')
                    self.c = pyfftw.empty_aligned(self.nfft, dtype='float32')
                    self.forward = pyfftw.FFTW(self.a, self.b)
                    self.backward = pyfftw.FFTW(self.b, self.c, 
                        direction='FFTW_BACKWARD')
                else:
                    self.a = pyfftw.empty_aligned([self.nfft,self.D], dtype='float32')
                    self.b = pyfftw.empty_aligned([self.nbin,self.D], dtype='complex64')
                    self.c = pyfftw.empty_aligned([self.nfft,self.D], dtype='float32')
                    self.forward = pyfftw.FFTW(self.a, self.b, axes=(0, ))
                    self.backward = pyfftw.FFTW(self.b, self.c, axes=(0, ), 
                        direction='FFTW_BACKWARD')
            else: 
                warnings.warn("Could not import pyfftw wrapper for fftw functions. Using numpy's rfft instead.")
                self.transform = 'numpy'
        elif transform == 'mkl':
            if mkl_available: 
                import mkl_fft
                self.transform = 'mkl'
            else:
                warnings.warn("Could not import mkl wrapper. Using numpy's rfft instead.")
                self.transform = 'numpy'
        else:
            self.transform = 'numpy'


    def analysis(self, x):
        """
        Perform frequency analysis of a real input using DFT.

        Parameters
        ----------
        x: numpy array (float32)
            Real input signal in time domain. Must be of size (N,D).
        plot_spec: bool
            Whether or not to plot frequency magnitude spectrum.

        Returns
        -------
        Frequency spectrum, numpy array of size (N/2+1,D) and of type complex64.
        """

        # check for valid input
        if self.D!=1:
            if x.shape!=(self.nfft,self.D):
                raise ValueError('Invalid input dimensions.')
        elif self.D==1:
            if x.ndim!=1 and x.shape[0]!=self.nfft:
                raise ValueError('Invalid input dimensions.')
        # apply window if needed
        if self.analysis_window is not None:
            try:
                np.multiply(self.analysis_window, x, x)
            except:
                self.analysis_window = self.analysis_window.astype(x.dtype, 
                    copy=False)
                np.multiply(self.analysis_window, x, x)
        # apply DFT
        if self.transform == 'fftw':
            self.a[:,] = x
            self.X[:,] = self.forward()
        elif self.transform == 'mkl':
            self.X[:,] = mkl_fft.rfft(x,axis=0)
        else:
            self.X[:,] = rfft(x,axis=0)

        return self.X


    def synthesis(self, X=None):
        """
        Perform time synthesis of frequency domain to real signal using the 
        inverse DFT.

        Parameters
        ----------
        X: numpy array (complex64)
            Real input signal in time domain. Must be of size (N/2+1,D).
            Default is previously computed DFT.
        plot_time: bool
            Whether or not to plot time waveform.

        Returns
        -------
        Time domain signal, numpy array of size (N,D) and of type float32.
        """

        # check for valid input
        if X is not None:
            if self.D!=1:
                if X.shape!=(self.nbin,self.D):
                    raise ValueError('Invalid input dimensions.')
            elif self.D==1:
                if X.ndim!=1 and X.shape[0]!=self.nbin:
                    raise ValueError('Invalid input dimensions.')
            self.X[:,] = X
        # inverse DFT
        if self.transform == 'fftw':
            self.b[:] = self.X
            self.x[:,] = self.backward()
        elif self.transform == 'mkl':
            self.x[:,] = mkl_fft.irfft(self.X,axis=0)
        else:
            self.x[:,] = irfft(self.X, axis=0)
        # apply window if needed
        if self.synthesis_window is not None:
            print('orly?')
            np.multiply(self.synthesis_window, self.x[:,], self.x[:,])

        return self.x

